@@ -1,4 +1,5 @@
 import datetime as dt
+from typing import ClassVar
 
 import pandas as pd
 import pandera as pa
@@ -8,7 +9,7 @@ from pandera.typing import DataFrame, Index, Series
 class JsonSchemaOut(pa.SchemaModel):
     class Config:
         to_format = "json"
-        to_format_kwargs = {"orient": "table"}
+        to_format_kwargs: ClassVar[dict[str, str]] = {"orient": "table"}
         coerce = True
 
 
@@ -60,10 +61,10 @@ MarketDatesDataFrame = DataFrame[MarketDatesSchema]
 
 class BusDataSchema(JsonSchemaOut):
     node: Index[str] = pa.Field(unique=True)
-    name: Series[str] = pa.Field()
     busname: Series[str] = pa.Field()
+    substation: Series[str] = pa.Field()
     un: Series[float] = pa.Field()
-    station: Series[str] = pa.Field()
+    substation_mrid: Series[str] = pa.Field()
     bidzone: Series[str] = pa.Field(nullable=True)
     sv_voltage: Series[float] = pa.Field(nullable=True)
     island: Series[str] = pa.Field()
@@ -75,12 +76,13 @@ BusDataFrame = DataFrame[BusDataSchema]
 
 class LoadsSchema(NamedResourceSchema):
     node: Series[str] = pa.Field()
-    station: Series[str] = pa.Field()
+    substation_mrid: Series[str] = pa.Field()
     bidzone: Series[str] = pa.Field(nullable=True)
     status: Series[bool] = pa.Field()
     p: Series[float] = pa.Field(nullable=True)
     q: Series[float] = pa.Field(nullable=True)
     station_group: Series[str] = pa.Field(nullable=True)
+    connectivity_node: Series[str] = pa.Field()
 
 
 LoadsDataFrame = DataFrame[LoadsSchema]
@@ -97,19 +99,19 @@ WindGeneratingUnitsDataFrame = DataFrame[WindGeneratingUnitsSchema]
 
 
 class SynchronousMachinesSchema(NamedMarketResourceSchema):
-    allocationmax: Series[float] = pa.Field(nullable=True)
     node: Series[str] = pa.Field()
     status: Series[bool] = pa.Field()
     station_group: Series[str] = pa.Field(nullable=True)
     station_group_name: Series[str] = pa.Field(nullable=True)
-    station: Series[str] = pa.Field()
-    maxP: Series[float] = pa.Field(nullable=True)
-    minP: Series[float] = pa.Field(nullable=True)
+    substation_mrid: Series[str] = pa.Field()
+    maxP: Series[float] = pa.Field(nullable=True)  # noqa N815
+    minP: Series[float] = pa.Field(nullable=True)  # noqa N815
     MO: Series[float] = pa.Field(nullable=True)
     bidzone: Series[str] = pa.Field(nullable=True)
     sn: Series[float] = pa.Field()
     p: Series[float] = pa.Field(nullable=True)
     q: Series[float] = pa.Field(nullable=True)
+    connectivity_node: Series[str] = pa.Field()
 
 
 SynchronousMachinesDataFrame = DataFrame[SynchronousMachinesSchema]
@@ -144,9 +146,12 @@ ExchangeDataFrame = DataFrame[ExchangeSchema]
 
 class ConvertersSchema(NamedResourceSchema):
     alias: Series[str] = pa.Field(nullable=True)
-    station: Series[str] = pa.Field()
+    substation_mrid: Series[str] = pa.Field()
     status: Series[bool] = pa.Field()
     node: Series[str] = pa.Field()
+    p: Series[float] = pa.Field()
+    q: Series[float] = pa.Field()
+    connectivity_node: Series[str] = pa.Field()
 
 
 ConvertersDataFrame = DataFrame[ConvertersSchema]
@@ -183,6 +188,8 @@ class BranchComponentSchema(NamedResourceSchema):
     status: Series[bool] = pa.Field()
     un: Series[float] = pa.Field()
     x: Series[float] = pa.Field()
+    connectivity_node_1: Series[str] = pa.Field()
+    connectivity_node_2: Series[str] = pa.Field()
 
 
 BranchComponentDataFrame = DataFrame[BranchComponentSchema]
@@ -289,8 +296,8 @@ StationGroupCodeNameDataFrame = DataFrame[StationGroupCodeNameSchema]
 
 
 class HVDCBidzonesSchema(MridResourceSchema):
-    from_area: Series[str] = pa.Field()
-    to_area: Series[str] = pa.Field()
+    bidzone_1: Series[str] = pa.Field()
+    bidzone_2: Series[str] = pa.Field()
 
 
 HVDCBidzonesDataFrame = DataFrame[HVDCBidzonesSchema]
@@ -303,3 +310,44 @@ class TransformerWindingsSchema(JsonSchemaOut):
 
 
 TransformerWindingsDataFrame = DataFrame[TransformerWindingsSchema]
+
+
+class SvInjectionSchema(JsonSchemaOut):
+    node: Series[str] = pa.Field(unique=True)
+    p: Series[float] = pa.Field()
+    q: Series[float] = pa.Field()
+
+
+SvInjectionDataFrame = DataFrame[SvInjectionSchema]
+
+
+class RASEquipmentSchema(JsonSchemaOut):
+    mrid: Series[str] = pa.Field(unique=True)
+    equipment_mrid: Series[str] = pa.Field()
+    name: Series[str] = pa.Field()
+
+
+RASEquipmentDataFrame = DataFrame[RASEquipmentSchema]
+
+
+class Switches(JsonSchemaOut):
+    mrid: Index[str] = pa.Field(unique=True)
+    is_open: Series[bool] = pa.Field()
+    equipment_type: Series[str] = pa.Field()
+    connectivity_node_1: Series[str] = pa.Field()
+    connectivity_node_2: Series[str] = pa.Field()
+
+
+SwitchesDataFrame = DataFrame[Switches]
+
+
+class ConnectivityNode(JsonSchemaOut):
+    mrid: Index[str] = pa.Field(unique=True)
+    container: Series[str] = pa.Field()
+    container_name: Series[str] = pa.Field()
+    un: Series[float] = pa.Field(nullable=True)
+    bidzone: Series[str] = pa.Field(nullable=True)
+    container_type: Series[str] = pa.Field()
+
+
+ConnectivityNodeDataFrame = DataFrame[ConnectivityNode]
