@@ -1,29 +1,33 @@
 """Graphdb CIM sparql client"""
+from __future__ import annotations
 
 import json
 import os
+from collections.abc import Callable
 from dataclasses import dataclass, field
 from enum import auto
+from http import HTTPStatus
 from pathlib import Path
-from typing import Dict, List, Optional, Tuple, TypedDict, Union
+from typing import TYPE_CHECKING, TypedDict, TypeVar
 
 import httpx
 import pandas as pd
-import requests
+import tenacity
 from SPARQLWrapper import JSON, POST, SPARQLWrapper
 from strenum import StrEnum
 
 from cimsparql.async_sparql_wrapper import AsyncSparqlWrapper
+from cimsparql.sparql_result_json import SparqlResultJson
 from cimsparql.url import service, service_blazegraph
 
 
 class SparqlResult(TypedDict):
-    cols: List[str]
-    data: List[Dict[str, Dict[str, str]]]
-    out: List[Dict[str, str]]
+    cols: list[str]
+    data: list[dict[str, dict[str, str]]]
+    out: list[dict[str, str]]
 
 
-def data_row(cols: List[str], rows: List[Dict[str, Dict[str, str]]]) -> Dict[str, str]:
+def data_row(cols: list[str], rows: list[dict[str, dict[str, str]]]) -> dict[str, str]:
     """Get a sample row for extraction of data types
 
     Args:
@@ -48,7 +52,7 @@ class RestApi(StrEnum):
     DIRECT_SPARQL_ENDPOINT = auto()
 
 
-def parse_namespaces_rdf4j(response: requests.Response) -> Dict[str, str]:
+def parse_namespaces_rdf4j(response: httpx.Response) -> dict[str, str]:
     prefixes = {}
     for line in response.text.split()[1:]:
         prefix, uri = line.split(",")
@@ -62,20 +66,23 @@ class ServiceConfig:
     protocol: str = "https"
     server: str = field(default=os.getenv("GRAPHDB_SERVER", "127.0.0.1:7200"))
     path: str = ""
-    user: str = field(default=os.getenv("GRAPHDB_USER"))
-    passwd: str = field(default=os.getenv("GRAPHDB_USER_PASSWD"))
-    rest_api: RestApi = field(default=os.getenv("SPARQL_REST_API", RestApi.RDF4J))
-    ca_bundle: Optional[str] = field(default=None)
+    user: str | None = field(default=os.getenv("GRAPHDB_USER"))
+    passwd: str | None = field(default=os.getenv("GRAPHDB_USER_PASSWD"))
+    rest_api: RestApi = field(default=RestApi(os.getenv("SPARQL_REST_API", "RDF4J")))
+    ca_bundle: str | None = field(default=None)
 
     # Parameters for rest api
     # https://rdf4j.org/documentation/reference/rest-api/
     distinct: bool = False
     infer: bool = False
-    limit: Optional[int] = None
-    offset: Optional[int] = None
-    timeout: Optional[int] = None
+    limit: int | None = None
+    offset: int | None = None
+    timeout: int | None = None
+    num_retries: int = 0
+    max_delay_seconds: int = 60
+    validate: bool = False
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         if self.rest_api not in RestApi:
             raise ValueError(f"rest_api must be one of {RestApi}")
 
@@ -88,7 +95,7 @@ class ServiceConfig:
         return service(self.repo, self.server, self.protocol, self.path)
 
     @property
-    def parameters(self) -> Dict[str, Union[bool, int]]:
+    def parameters(self) -> dict[str, bool | int | None]:
         return {
             "distinct": self.distinct,
             "infer": self.infer,
@@ -96,6 +103,10 @@ class ServiceConfig:
             "offset": self.offset,
         }
 
+    @property
+    def auth(self) -> httpx.BasicAuth | None:
+        return httpx.BasicAuth(self.user, self.passwd) if self.user and self.passwd else None
+
 
 # Available formats from RDF4J API
 # https://rdf4j.org/documentation/reference/rest-api/
@@ -114,41 +125,62 @@ MIME_TYPE_RDF_FORMATS = {
 
 UPLOAD_END_POINT = {RestApi.RDF4J: "/statements", RestApi.BLAZEGRAPH: ""}
 
+if TYPE_CHECKING:
+    from typing import Concatenate, ParamSpec
 
-def require_rdf4j(f):
-    def wrapper(*args):
-        self = args[0]
-        if self.service_cfg.rest_api != RestApi.RDF4J:
+    P = ParamSpec("P")
+    T = TypeVar("T")
+
+    RDF4J_DECORATED = Callable[Concatenate[GraphDBClient, P], T]
+
+
+def require_rdf4j(f: RDF4J_DECORATED) -> Callable[RDF4J_DECORATED]:
+    def wrapper(cli: GraphDBClient, *args: P.args, **kwargs: P.kwargs) -> T:
+        if cli.service_cfg.rest_api != RestApi.RDF4J:
             raise NotImplementedError("Function only implemented for RDF4J")
-        return f(*args)
+        return f(cli, *args, **kwargs)
 
     return wrapper
 
 
 class GraphDBClient:
-    def __init__(self, service_cfg: Optional[ServiceConfig] = None) -> None:
-        """GraphDB client
+    """GraphDB client for sending sparql queries to GraphDB server
 
-        Args:
-           service: string with url to graphdb repository. See cimsparql.url.service
-           mapper: GraphDBClient with the mapper (Default to self).
-           infer: deduce further knowledge based on existing RDF data and a formal set of
-           sameas: map same concepts from two or more datasets
-        """
+    Args:
+        service_cfg: Service configuration (see ServiceConfig)
+        custom_headers: Added to SPARQLWrapper using addCustomHttpHeader
+
+    Example:
+    >>> from cimsparql.graphdb import GraphDBClient
+    >>> gdbc = GraphDBClient()
+    >>> query = 'select * where { ?subject ?predicate ?object } limit 10'
+    >>> df, row = gdbc.get_table(query)xs
+
+    Where row is the output of graphdb.data_row
+    """
+
+    def __init__(
+        self,
+        service_cfg: ServiceConfig | None = None,
+        custom_headers: dict[str, str] | None = None,
+    ) -> None:
         self.service_cfg = service_cfg or ServiceConfig()
-        self.sparql = SPARQLWrapper(self.service_cfg.url)
-        self._init_sparql_wrapper()
-        self._prefixes = None
-
-    def _init_sparql_wrapper(self):
+        self.sparql = self.create_sparql_wrapper()
         self.sparql.setReturnFormat(JSON)
         self.sparql.setMethod(POST)
         self.sparql.setCredentials(self.service_cfg.user, self.service_cfg.passwd)
         if self.service_cfg.timeout:
             self.sparql.setTimeout(self.service_cfg.timeout)
         self._update_sparql_parameters()
+        if custom_headers:
+            for name, value in custom_headers.items():
+                self.sparql.addCustomHttpHeader(name, value)
+        self._prefixes = None
 
-    def _update_sparql_parameters(self):
+    def create_sparql_wrapper(self) -> SPARQLWrapper:
+        return SPARQLWrapper(self.service_cfg.url)
+
+    def _update_sparql_parameters(self) -> None:
         for key, value in self.service_cfg.parameters.items():
             if value is not None:
                 self.set_parameter(key, str(value))
@@ -162,12 +194,12 @@ class GraphDBClient:
         self.sparql.endpoint = self.service_cfg.url
 
     @property
-    def prefixes(self) -> Dict[str, str]:
+    def prefixes(self) -> dict[str, str]:
         if self._prefixes is None:
             self._prefixes = self.get_prefixes()
         return self._prefixes
 
-    def update_prefixes(self, pref: Dict[str, str]):
+    def update_prefixes(self, pref: dict[str, str]) -> None:
         """
         Update prefixes from a dict
         """
@@ -176,40 +208,46 @@ class GraphDBClient:
     def __str__(self) -> str:
         return f"<GraphDBClient object, service: {self.service_cfg.url}>"
 
-    def _prep_query(self, query: str):
+    def _prep_query(self, query: str) -> None:
         self.sparql.setQuery(query)
         self._update_sparql_parameters()
 
-    def _process_result(self, results: dict) -> dict:
-        cols = results["head"]["vars"]
-        data = results["results"]["bindings"]
-        out = [{c: row.get(c, {}).get("value") for c in cols} for row in data]
+    @staticmethod
+    def _process_result(results: SparqlResultJson) -> dict:
+        cols = results.head.variables
+        data = results.results.bindings
+        out = [{c: row[c].value if c in row else None for c in cols} for row in data]
         return {"out": out, "cols": cols, "data": data}
 
     def _exec_query(self, query: str) -> SparqlResult:
         self._prep_query(query)
-        results = self.sparql.queryAndConvert()
-        return self._process_result(results)
 
-    def exec_query(self, query: str) -> List[Dict[str, str]]:
+        for attempt in tenacity.Retrying(
+            stop=tenacity.stop_after_attempt(self.service_cfg.num_retries + 1),
+            wait=tenacity.wait_exponential(max=self.service_cfg.max_delay_seconds),
+        ):
+            with attempt:
+                results = self.sparql.queryAndConvert()
+
+        sparql_result = SparqlResultJson(**results)
+        if self.service_cfg.validate:
+            sparql_result.validate_column_consistency()
+        return self._process_result(sparql_result)
+
+    def exec_query(self, query: str) -> list[dict[str, str]]:
         out = self._exec_query(query)
         return out["out"]
 
-    def _convert_query_result_to_df(self, res: SparqlResult) -> Tuple[pd.DataFrame, Dict[str, str]]:
+    def _convert_query_result_to_df(self, res: SparqlResult) -> tuple[pd.DataFrame, dict[str, str]]:
         df = pd.DataFrame(res["out"]) if len(res["out"]) else pd.DataFrame(columns=res["cols"])
         return df, data_row(res["cols"], res["data"])
 
-    def get_table(self, query: str) -> Tuple[pd.DataFrame, Dict[str, str]]:
-        """
+    def get_table(self, query: str) -> tuple[pd.DataFrame, dict[str, str]]:
+        """Get result from sparql query as a pandas dataframe
+
         Args:
            query: to sparql server
            limit: limit number of resulting rows
-        Example:
-           >>> from cimsparql.graphdb import GraphDBClient
-           >>> from cimsparql.url import service
-           >>> gdbc = GraphDBClient(service('LATEST'))
-           >>> query = 'select * where { ?subject ?predicate ?object }'
-           >>> gdbc.get_table(query, limit=10)
         """
         res = self._exec_query(query)
         return self._convert_query_result_to_df(res)
@@ -219,7 +257,7 @@ class GraphDBClient:
         """Identify empty GraphDB repo"""
         return self.get_table("select * where {?s ?p ?o} limit 1")[0].empty
 
-    def get_prefixes(self) -> Dict[str, str]:
+    def get_prefixes(self) -> dict[str, str]:
         prefixes = default_namespaces()
 
         if self.service_cfg.rest_api in (RestApi.BLAZEGRAPH, RestApi.DIRECT_SPARQL_ENDPOINT):
@@ -227,9 +265,12 @@ class GraphDBClient:
             # via `update_prefixes`. By default we load a pre-defined set of prefixes
             return prefixes
 
-        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
-        response = requests.get(self.service_cfg.url + "/namespaces", auth=auth)
-        if response.ok:
+        response = httpx.get(
+            self.service_cfg.url + "/namespaces",
+            auth=self.service_cfg.auth,
+            headers=self.sparql.customHttpHeaders,
+        )
+        if response.status_code == HTTPStatus.OK:
             prefixes.update(parse_namespaces_rdf4j(response))
             return prefixes
         msg = (
@@ -237,15 +278,17 @@ class GraphDBClient:
             "Verify that user and password are correctly set in the "
             "GRAPHDB_USER and GRAPHDB_USER_PASSWD environment variable"
         )
-        raise RuntimeError(f"{msg} Status code: {response.status_code} Reason: {response.reason}")
+        raise RuntimeError(
+            f"{msg} Status code: {response.status_code} Reason: {response.reason_phrase}"
+        )
 
     def delete_repo(self) -> None:
         endpoint = delete_repo_endpoint(self.service_cfg)
-        response = requests.delete(endpoint)
+        response = httpx.delete(endpoint)
         response.raise_for_status()
 
     def upload_rdf(
-        self, content: Union[Path, bytes], rdf_format: str, params: Optional[Dict[str, str]] = None
+        self, content: Path | bytes, rdf_format: str, params: dict[str, str] | None = None
     ) -> None:
         """
         Upload data in RDF format to a srevice
@@ -256,14 +299,16 @@ class GraphDBClient:
                 doc for a complete list of available options)
             params: Additional parameters passed to the post request
         """
-        xml_content = content
-        if isinstance(content, Path):
-            with open(content, "rb") as infile:
-                xml_content = infile.read()
 
-        response = requests.post(
+        def read_xml_content(file: Path) -> bytes:
+            with open(file, "rb") as infile:
+                return infile.read()
+
+        xml_content = read_xml_content(content) if isinstance(content, Path) else content
+
+        response = httpx.post(
             self.service_cfg.url + UPLOAD_END_POINT[self.service_cfg.rest_api],
-            data=xml_content,
+            content=xml_content,
             params=params,
             headers={"Content-Type": MIME_TYPE_RDF_FORMATS[rdf_format]},
         )
@@ -273,30 +318,30 @@ class GraphDBClient:
         """
         Function that passes a query via a post API call
         """
-        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
-        response = requests.post(
+        response = httpx.post(
             self.service_cfg.url + UPLOAD_END_POINT[self.service_cfg.rest_api],
             data={"update": query},
-            headers={"Content-Type": "application/x-www-form-urlencoded"},
-            auth=auth,
+            headers=self.sparql.customHttpHeaders
+            | {"Content-Type": "application/x-www-form-urlencoded"},
+            auth=self.service_cfg.auth,
         )
         response.raise_for_status()
 
     @require_rdf4j
     def set_namespace(self, prefix: str, value: str) -> None:
-        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
-        response = requests.put(
+        response = httpx.put(
             self.service_cfg.url + f"/namespaces/{prefix}",
-            data=value,
+            content=value,
             headers={"Content-Type": "text/plain"},
-            auth=auth,
+            auth=self.service_cfg.auth,
         )
         response.raise_for_status()
 
     @require_rdf4j
     def get_namespace(self, prefix: str) -> str:
-        auth = requests.auth.HTTPBasicAuth(self.service_cfg.user, self.service_cfg.passwd)
-        response = requests.get(self.service_cfg.url + f"/namespaces/{prefix}", auth=auth)
+        response = httpx.get(
+            self.service_cfg.url + f"/namespaces/{prefix}", auth=self.service_cfg.auth
+        )
         response.raise_for_status()
         return response.text
 
@@ -310,7 +355,7 @@ class RepoInfo:
     writable: bool
 
 
-async def repos(service_cfg: Optional[ServiceConfig] = None) -> List[RepoInfo]:
+async def repos(service_cfg: ServiceConfig | None = None) -> list[RepoInfo]:
     """
     List available repositories
     """
@@ -327,8 +372,8 @@ async def repos(service_cfg: Optional[ServiceConfig] = None) -> List[RepoInfo]:
         response = await client.get(url, auth=auth, headers={"Accept": "application/json"})
     response.raise_for_status()
 
-    def _repo_info(repo):
-        def get(key, default=""):
+    def _repo_info(repo: dict[str, str]) -> RepoInfo:
+        def get(key: str, default: str = "") -> str:
             return repo.get(key, {}).get("value", default)
 
         uri = get("uri")
@@ -355,9 +400,9 @@ def new_repo(
             Otherwise an error is raised if a repository with the same name already exists
         protocol: Default https
     """
-    ignored_errors = {409} if allow_exist else set()
+    ignored_errors = {HTTPStatus.CONFLICT} if allow_exist else set()
     url = service(repo, server, protocol)
-    response = requests.put(url, data=config, headers={"Content-Type": "text/turtle"})
+    response = httpx.put(url, content=config, headers={"Content-Type": "text/turtle"})
     if response.status_code not in ignored_errors:
         response.raise_for_status()
 
@@ -368,8 +413,8 @@ def new_repo_blazegraph(url: str, repo: str, protocol: str = "https") -> GraphDB
     template = confpath() / "blazegraph_repo_config.xml"
     config = config_bytes_from_template(template, {"repo": repo})
 
-    response = requests.post(
-        f"{protocol}://{url}", data=config, headers={"Content-type": "application/xml"}
+    response = httpx.post(
+        f"{protocol}://{url}", content=config, headers={"Content-type": "application/xml"}
     )
     response.raise_for_status()
     client = GraphDBClient(ServiceConfig(repo, protocol, url, rest_api=RestApi.BLAZEGRAPH))
@@ -377,7 +422,7 @@ def new_repo_blazegraph(url: str, repo: str, protocol: str = "https") -> GraphDB
 
 
 def config_bytes_from_template(
-    template: Path, params: Dict[str, str], encoding: str = "utf8"
+    template: Path, params: dict[str, str], encoding: str = "utf-8"
 ) -> bytes:
     """
     Replace value in template file with items in params
@@ -401,7 +446,7 @@ def confpath() -> Path:
     return Path(__file__).parent.parent / "pkg_data"
 
 
-def default_namespaces() -> Dict[str, str]:
+def default_namespaces() -> dict[str, str]:
     with open(confpath() / "namespaces.json") as infile:
         return json.load(infile)
 
@@ -414,31 +459,45 @@ def delete_repo_endpoint(config: ServiceConfig) -> str:
 
 
 class AsyncGraphDBClient(GraphDBClient):
-    def __init__(self, service_cfg: Optional[ServiceConfig] = None) -> None:
-        super().__init__(service_cfg)
-        self.sparql = AsyncSparqlWrapper(self.service_cfg.url, self.service_cfg.ca_bundle)
-        self._init_sparql_wrapper()
+    """Asynchronous GraphDB client for sending sparql queries to GraphDB server
+
+    Args:
+        service_cfg: Service configuration (see ServiceConfig)
+        custom_headers: Added to SPARQLWrapper using addCustomHttpHeader
+
+    Example:
+    >>> import asyncio
+    >>> from cimsparql.graphdb import AsyncGraphDBClient
+    >>> gdbc = AsyncGraphDBClient()
+    >>> query = 'select * where { ?subject ?predicate ?object } limit 10'
+    >>> df, row = await gdbc.get_table(query)
+
+    Where row is the output of graphdb.data_row
+    """
+
+    def create_sparql_wrapper(self) -> AsyncSparqlWrapper:
+        return AsyncSparqlWrapper(
+            self.service_cfg.url,
+            ca_bundle=self.service_cfg.ca_bundle,
+            num_retries=self.service_cfg.num_retries,
+            max_delay_seconds=self.service_cfg.max_delay_seconds,
+        )
 
     async def _exec_query(self, query: str) -> SparqlResult:
         self._prep_query(query)
-        results = await self.sparql.queryAndConvert()
+        results = await self.sparql.query_and_convert()
         return self._process_result(results)
 
-    async def exec_query(self, query: str) -> List[Dict[str, str]]:
+    async def exec_query(self, query: str) -> list[dict[str, str]]:
         out = await self._exec_query(query)
         return out["out"]
 
-    async def get_table(self, query: str) -> Tuple[pd.DataFrame, Dict[str, str]]:
-        """
+    async def get_table(self, query: str) -> tuple[pd.DataFrame, dict[str, str]]:
+        """Get result from sparql query as a pandas dataframe
+
         Args:
            query: to sparql server
            limit: limit number of resulting rows
-        Example:
-           >>> from cimsparql.graphdb import GraphDBClient
-           >>> from cimsparql.url import service
-           >>> gdbc = GraphDBClient(service('LATEST'))
-           >>> query = 'select * where { ?subject ?predicate ?object }'
-           >>> gdbc.get_table(query, limit=10)
         """
         res = await self._exec_query(query)
         return self._convert_query_result_to_df(res)
