@@ -3,7 +3,7 @@ from string import Template
 
 
 def _read_template(filename: pathlib.Path) -> Template:
-    with open(filename, "r") as file:
+    with open(filename) as file:
         return Template(file.read())
 
 
@@ -15,6 +15,7 @@ BORDERS_QUERY = _read_template(sparql_folder / "borders.sparql")
 BRANCH_NODE_WITHDRAW_QUERY = _read_template(sparql_folder / "branch_node_withdraw.sparql")
 BUS_DATA_QUERY = _read_template(sparql_folder / "bus.sparql")
 CONNECTIONS_QUERY = _read_template(sparql_folder / "connections.sparql")
+CONNECTIVITY_NODES_QUERY = _read_template(sparql_folder / "connectivity_nodes.sparql")
 CONVERTERS_QUERY = _read_template(sparql_folder / "converters.sparql")
 COORDINATES_QUERY = _read_template(sparql_folder / "coordinates.sparql")
 DC_ACTIVE_POWER_FLOW_QUERY = _read_template(sparql_folder / "dc_active_power_flow.sparql")
@@ -25,7 +26,7 @@ HVDC_CONVERTER_BIDZONES = _read_template(sparql_folder / "converter_hvdc_bidzone
 LOADS_QUERY = _read_template(sparql_folder / "loads.sparql")
 MARKET_DATES_QUERY = _read_template(sparql_folder / "market_dates.sparql")
 POWER_FLOW_QUERY = _read_template(sparql_folder / "power_flow.sparql")
-PROTECTIVE_ACTION_QUERY = _read_template(sparql_folder / "protective_action.sparql")
+RAS_EQUIPMENT_QUERY = _read_template(sparql_folder / "ras_equipment.sparql")
 REGIONS_QUERY = _read_template(sparql_folder / "regions.sparql")
 SERIES_COMPENSATORS_QUERY = _read_template(sparql_folder / "series_compensators.sparql")
 STATION_GROUP_CODE_NAME_QUERY = _read_template(
@@ -33,7 +34,9 @@ STATION_GROUP_CODE_NAME_QUERY = _read_template(
 )
 SUBSTATION_VOLTAGE_LEVEL_QUERY = _read_template(sparql_folder / "substation_voltage_level.sparql")
 SV_BRANCH_QUERY = _read_template(sparql_folder / "sv_branch.sparql")
+SV_INJECTION_QUERY = _read_template(sparql_folder / "sv_injection.sparql")
 SYNCHRONOUS_MACHINES_QUERY = _read_template(sparql_folder / "synchronous_machines.sparql")
+SWITCHES_QUERY = _read_template(sparql_folder / "switches.sparql")
 THREE_WINDING_DUMMY_NODES_QUERY = _read_template(sparql_folder / "three_winding_dummy_nodes.sparql")
 THREE_WINDING_LOSS_QUERY = _read_template(sparql_folder / "three_winding_loss.sparql")
 THREE_WINDING_QUERY = _read_template(sparql_folder / "three_winding.sparql")
@@ -46,3 +49,4 @@ TWO_WINDING_QUERY = _read_template(sparql_folder / "two_winding_transformer.spar
 TYPE_MAPPER_QUERY = _read_template(sparql_folder / "type_mapper.sparql")
 WIND_GENERATING_UNITS_QUERY = _read_template(sparql_folder / "wind_generating_units.sparql")
 TRANSFORMER_WINDINGS_QUERY = _read_template(sparql_folder / "transformer_windings.sparql")
+WINDING = _read_template(sparql_folder / "winding.sparql")
